/*
 * Ability.h, part of VCMI engine
 *
 * Authors: listed in file AUTHORS in main folder
 *
 * License: GNU General Public License v2.0 or later
 * Full text of license available in license.txt file, in main folder
 *
 */

VCMI_LIB_NAMESPACE_BEGIN

namespace abilities
{

class DLL_LINKAGE Ability
{

};

}

VCMI_LIB_NAMESPACE_END
